#
# GDisPAL ("Genetic Distance Patterns Across Landscapes"), June 2014
# Based on the method of Miller (2005)
# Script created by Simon Dellicour (Evolutionary Biology & Ecology - ULB)
#
# GDisPAL(rast, coordinates, distances, a, nberOfCores):
#
#   - rast: raster on which the interpollation will be performed.
#   - coordinates: [X Y], with X the vector of longitude coordinates 
#       (decimal degrees) and Y the vector of latitude coordinates 
#       (decimal degrees) of each sampled locality
#   - distances: square matrix containing the pairwise distance 
#       indices between all the sampled localities
#     N.B.: the order of the sampled locilities coordinates in the
#       "coordinates" matrix must correspond to the order of the pairwise 
#       distances in the "distances" matrix
#   - a: the distance weighting value (see Miller 2005 and AIS manual for 
#       further details and a discussion about the impact of this 
#       weighting parameter)
#   - nberOfCores: number of cores (CPU) available to parallelise the script
#

GDisPAL = function(rast, coordinates, distances, a, nberOfCores)
{    
	registerDoMC(cores=nberOfCores)
   	rast1 = rast; rast2 = rast
    rast1[] = 0

	# Delaunay triangulation:
    triangulation = delaunayn(coordinates)
    size_triangulation = dim(triangulation)
    midpoints = matrix(0,nrow=(3*size_triangulation[1]),ncol=5)
    edges = matrix(-1,nrow=(3*size_triangulation[1]),ncol=2)
    c = 0; edge = 1
    for (i in 1:size_triangulation[1])
    	{
    		x1 = coordinates[triangulation[i,1],1]
			y1 = coordinates[triangulation[i,1],2]
			x2 = coordinates[triangulation[i,2],1]
			y2 = coordinates[triangulation[i,2],2]
			x3 = coordinates[triangulation[i,3],1]
			y3 = coordinates[triangulation[i,3],2]
			
			new_edge = 1 # true
			pop1 = min(triangulation[i,1],triangulation[i,2])
			pop2 = max(triangulation[i,1],triangulation[i,2])
			for (j in 1:size_triangulation[1])
				{
					if(pop1 == edges[j,1])
						{
							if(pop2 == edges[j,2])
								{
									new_edge = -1 # false
								}
						}	
				}
			if(new_edge == 1)
				{
					c = c+1;
					midpoints[c,1] = triangulation[i,1]
					midpoints[c,2] = triangulation[i,2]
					midpoints[c,3] = (x1+x2)/2
					midpoints[c,4] = (y1+y2)/2
					midpoints[c,5] = distances[triangulation[i,1],triangulation[i,2]]
					edges[edge,1] = pop1; edges[edge,2] = pop2
					edge = edge+1
				}
			
			new_edge = 1 # true
			pop1 = min(triangulation[i,2],triangulation[i,3]) 
			pop2 = max(triangulation[i,2],triangulation[i,3]) 
			for (j in 1:size_triangulation[1])
				{
					if(pop1 == edges[j,1])
						{
							if(pop2 == edges[j,2])
								{
									new_edge = -1 # false
								}
						}	
				}
			if(new_edge == 1)
				{	
					c = c+1
					midpoints[c,1] = triangulation[i,2]
					midpoints[c,2] = triangulation[i,3]
					midpoints[c,3] = (x2+x3)/2
					midpoints[c,4] = (y2+y3)/2
					midpoints[c,5] = distances[triangulation[i,2],triangulation[i,3]]
					edges[edge,1] = pop1; edges[edge,2] = pop2
					edge = edge+1
				}
				
			new_edge = 1; # true
			pop1 = min(triangulation[i,1],triangulation[i,3]); 
			pop2 = max(triangulation[i,1],triangulation[i,3]); 
			for (j in 1:size_triangulation[1])
				{
					if(pop1 == edges[j,1])
						{
							if(pop2 == edges[j,2])
								{
									new_edge = -1 # false
								}
						}	
				}
			if(new_edge == 1)
				{	
					c = c+1;
					midpoints[c,1] = triangulation[i,1]
					midpoints[c,2] = triangulation[i,3]
					midpoints[c,3] = (x1+x3)/2
					midpoints[c,4] = (y1+y3)/2
					midpoints[c,5] = distances[triangulation[i,1],triangulation[i,3]]
					edges[edge,1] = pop1; edges[edge,2] = pop2
					edge = edge+1
				}
    	}
    x_obs = midpoints[1:(edge-1),3]
    y_obs = midpoints[1:(edge-1),4]
    z_obs = midpoints[1:(edge-1),5]
    
	# Inverse-distance weighted interpolation:
    buffer = foreach(i = 1:length(rast1[])) %dopar%	{
    # for (i in 1:length(rast1[]))
    	# {
        	xy = xyFromCell(rast1,i)
            numerator = 0
            wi_sum = 0
            for (k in 1:length(z_obs))
            	{
            		wi = 0
            		if ((xy[1] == x_obs[k]) & (xy[2]==y_obs[k]))
            			{
            				wi = 1; counter = counter+1
            			}
            		else 
            			{
            				xy_obs = cbind(x_obs[k], y_obs[k])
							distance = rdist.earth(xy, xy_obs, miles=F, R=NULL)
							wi = (distance^(-(a/2)))
            				# wi = ((x_obs[k]-xy[1])^2+(y_obs[k]-xy[2])^2)^(-(a/2))
            			}                
            		wi_sum = wi_sum+wi
            		numerator = numerator+(wi*z_obs[k])
            	}
            z = numerator/wi_sum
            rast1[i] = z
    	}	
    for (i in 1:length(buffer))
    	{
    		rast1[i] = buffer[[i]]
    	}

	rast2 = rast1; rast2[is.na(rast[])] = NA
    return(rast2)
}
